import numpy as np


def generate_circuit(length, parent=None, mutate=0.3, parallel=0.2):
    i = 0
    elem = np.random.randint(0, length, size=length)
    l1 = ['R', 'C', 'W1/W2', 'A', 'E1/E2', 'G1/G2']
    out = ""
    if parent is None:
        while i < length:
            par = np.random.random()

            if par < parallel and i < length-1:
                out += "p("
                out += l1[elem[i]]
                out += ','
                i += 1
                out += l1[elem[i]]
                out += ')-'
                i += 1
            else:
                out += l1[elem[i]]
                out += "-"
                i += 1
        out = out[:-1]
        out += ""
        return out
    else:
        discrete_parts = parent.split('-')
        chance = np.random.rand(len(discrete_parts))
        for i in range(len(discrete_parts)):
            if chance[i] < mutate:
                if i > 0:
                    out = ""
                else:
                    out = ""
                par = np.random.random()
                if par < parallel:
                    out += "p("
                    out += l1[elem[i]]
                    out += ','
#                    i += 1
                    out += l1[elem[i]]
                    out += ')'
#                    i += 1
                else:
                    out += l1[elem[i]]
#                    out += "-"
                if i == len(discrete_parts)-1:
                    out += ""
                discrete_parts[i] = out
        to_out = []
        for i in discrete_parts:
            to_out.append(i)
            to_out.append('-')
        return ''.join(to_out[:-1])


def make_population(num, n, parent=None):
    gen = []
    if parent is None:
        for i in range(num):
            gen.append(generate_circuit(n, parent=parent))
    return gen



import numpy as np
import cmath


def s(series):
    """ sums elements in series

    Notes
    ---------
    .. math::
        Z = Z_1 + Z_2 + ... + Z_n

    """
    z = len(series[0])*[0 + 0*1j]
    for elem in series:
        z += elem
    return z


def p(parallel):
    """ adds elements in parallel

    Notes
    ---------
    .. math::

        Z = \\frac{1}{\\frac{1}{Z_1} + \\frac{1}{Z_2} + ... + \\frac{1}{Z_n}}

     """
    z = len(parallel[0])*[0 + 0*1j]
    for elem in parallel:
        z += 1/elem
    return 1/z


def R(p, f):
    """ defines a resistor

    Notes
    ---------
    .. math::

        Z = R

    """

    typeChecker(p, f, R.__name__, 1)
    return np.array(len(f)*[p[0]])


def C(p, f):
    """ defines a capacitor

    .. math::

        Z = \\frac{1}{C \\times j 2 \\pi f}

     """

    typeChecker(p, f, C.__name__, 1)
    omega = 2*np.pi*np.array(f)
    Cap = p[0]

    return 1.0/(Cap*1j*omega)


def W(p, f):
    """ defines a blocked boundary Finite-length Warburg Element

    Notes
    ---------
    .. math::
        Z = \\frac{R}{\\sqrt{ T \\times j 2 \\pi f}} \\coth{\\sqrt{T \\times j 2 \\pi f }}  # noqa: E501

    where :math:`R` = p[0] (Ohms) and
    :math:`T` = p[1] (sec) = :math:`\\frac{L^2}{D}`

    """

    typeChecker(p, f, W.__name__, 2)

    omega = 2*np.pi*np.array(f)

    Zw = np.vectorize(lambda y: p[0]/(np.sqrt(p[1]*1j*y) *
                                      cmath.tanh(np.sqrt(p[1]*1j*y))))

    return Zw(omega)


def A(p, f):
    """ defines a semi-infinite Warburg element

    """

    typeChecker(p, f, A.__name__, 1)
    omega = 2*np.pi*np.array(f)
    Aw = p[0]

    Zw = Aw*(1-1j)*np.sqrt(omega)

    return Zw


def E(p, f):
    """ defines a constant phase element

    Notes
    -----
    .. math::

        Z = \\frac{1}{Q \\times (j 2 \\pi f)^\\alpha}

    where :math:`Q` = p[0] and :math:`\\alpha` = p[1].
    """
    typeChecker(p, f, E.__name__, 2)
    omega = 2*np.pi*np.array(f)
    Q = p[0]
    alpha = p[1]

    return 1.0/(Q*(1j*omega)**alpha)


def G(p, f):
    """ defines a Gerischer Element

    Notes
    ---------
    .. math::

        Z = \\frac{1}{Y \\times \\sqrt{K + j 2 \\pi f }}

     """
    typeChecker(p, f, G.__name__, 2)
    omega = 2*np.pi*np.array(f)
    Z0 = p[0]
    k = p[1]

    return Z0/np.sqrt(k + 1j*omega)


def typeChecker(p, f, name, length):
    assert isinstance(p, list), \
        'in {}, input must be of type list'.format(name)
    for i in p:
        assert isinstance(i, (float, int, np.int32, np.float64)), \
            'in {}, value {} in {} is not a number'.format(name, i, p)
    for i in f:
        assert isinstance(i, (float, int, np.int32, np.float64)), \
            'in {}, value {} in {} is not a number'.format(name, i, f)
    assert len(p) == length, \
        'in {}, input list must be length {}'.format(name, length)
    return



from .circuits import CustomCircuit
import numpy as np


def rmse(a, b):
    """
    A function which calculates the root mean squared
    error between two vectors.

    Notes
    ---------
    .. math::

        RMSE = \\sqrt{\\frac{1}{n}(a-b)^2}
    """

    return(np.abs(np.sqrt(np.mean(np.square(a-b)))))


def measurementModel(frequencies, impedances, algorithm='SLSQP',
                     max_k=7, R_guess=0.1, C_guess=10):
    """ Runs a measurement model test for validating impedance data

    Iteratively add RC circuit elements until the error converges.
    If error does not converge, it indicates that the data doesn't meet
    standards for linearity.

    Notes
    ---------
    .. math::

        RMSE = R_0 + \\sum_{0}^{k} R_i || C_i

    Parameters
    ---------
    frequencies: np.ndarray
        A list of frequencies to test
    impedances: np.ndarray of complex numbers
        A list of values to match to
    max_k: int
        The maximum number of RC elements to fit
    initial_guess: np.ndarray
        Initial guesses for R and C elements
    """

    circuit = "R_0"
    initial_guess = [R_guess]
    error_list = []
    model_list = []

    print('--- Running Measurement Model ---')
    for i in range(max_k):
        circuit += "-p(R_{},C_{})".format((i+1) % 9, (i+1) % 9)
        initial_guess.append(R_guess)
        initial_guess.append(C_guess)
        measModel = CustomCircuit(initial_guess=initial_guess,
                                  circuit=circuit, algorithm=algorithm)

        measModel.fit(frequencies, impedances)
        initial_guess = list(measModel.parameters_)
        model_list.append(measModel)
        fit = measModel.predict(frequencies)
        error = rmse(impedances, fit)
        error_list.append(error)

        print('Num elements: {}  Error: {:.2e}'.format(i+1, error))

    return model_list, error_list



from impedance.validation import rmse
import numpy as np


def test_RMSE():
    a = np.array([2 + 4*1j, 3 + 2*1j])
    b = np.array([2 + 4*1j, 3 + 2*1j])

    assert rmse(a, b) == 0.0

    c = np.array([2 + 4*1j, 1 + 4*1j])
    d = np.array([4 + 2*1j, 3 + 2*1j])

    assert rmse(c, d) == 2*np.sqrt(2)



from impedance.circuit_elements import R, C, W, A, E, G, s, p  # noqa: F401
import numpy as np
from scipy.optimize import leastsq
from scipy.optimize import minimize


def rmse(a, b):
    """
    A function which calculates the root mean squared error
    between two vectors.

    Notes
    ---------
    .. math::

        RMSE = \\sqrt{\\frac{1}{n}(a-b)^2}
    """

    return(np.abs(np.sqrt(np.mean(np.square(a-b)))))


def circuit_fit(frequencies, impedances, circuit,
                initial_guess, algorithm='leastsq', bounds=None):

    """ Main function for fitting an equivalent circuit to data

    Parameters
    -----------------
    frequencies : numpy array
        Frequencies

    impedances : numpy array of dtype 'complex128'
        Impedances

    circuit : string
        string defining the equivalent circuit to be fit

    initial_guess : list of floats
        initial guesses for the fit parameters
    algorithm: string
        Name of algorithm to pass to scipy.optimize.minimize
        or to instantiate scipy.optimize.leastsq

    Returns
    ------------
    p_values : list of floats
        best fit parameters for specified equivalent circuit

    p_errors : list of floats
        error estimates for fit parameters

    Notes
    ---------
    Need to do a better job of handling errors in fitting.
    Currently, an error of -1 is returned.

    """

    circuit = circuit.replace('_', '')

    f = frequencies
    Z = impedances

    if algorithm == 'leastsq':
        p_values, covar, _, _, ier = leastsq(residuals, initial_guess,
                                             args=(Z, f, circuit),
                                             maxfev=100000, ftol=1E-13,
                                             full_output=True)
        p_error = []
        if ier in [1, 2, 3, 4] and covar is not None:
            s_sq = (residuals(p_values, Z, frequencies, circuit)**2).sum()
            p_cov = covar * s_sq/(len(Z) - len(p_values))
            for i, __ in enumerate(covar):
                p_error.append(np.absolute(p_cov[i][i])**0.5)
        else:
            p_error = None
    elif algorithm in ['SLSQP', 'L-BFGS-B', 'TNC']:
        if bounds is None:
            bounds = []
            p_string = [x for x in circuit if x not in 'ps(),-/']
            for i, (a, b) in enumerate(zip(p_string[::2], p_string[1::2])):
                if str(a+b) == "E2":
                    bounds.append((0, 1))
                else:
                    bounds.append((0, None))
        res = minimize(residualWrapper, initial_guess, args=(Z, f, circuit),
                       method=algorithm, bounds=bounds)
        p_values = res.x
        covar = None
        p_error = len(p_values)*[-1]

    else:
        res = minimize(residualWrapper, initial_guess, args=(Z, f, circuit),
                       method=algorithm)
        p_values = res.x
        covar = None
        p_error = None

    return p_values, p_error


def residualWrapper(param, Z, f, circuit):
    res = residuals(param, Z, f, circuit)
    return np.mean(np.square(res))


def residuals(param, Z, f, circuit):
    """ Calculates the residuals between a given circuit/parameters
    (fit) and `Z`/`f` (data). Minimized by scipy.leastsq()

    Parameters
    ----------
    param : array of floats
        parameters for evaluating the circuit

    Z : array of complex numbers
        impedance data being fit

    f : array of floats
        frequencies to evaluate

    circuit : str
        string defining the circuit

    Returns
    -------
    residual : ndarray
        returns array of size 2*len(f) with both real and imaginary residuals
    """
    err = Z - computeCircuit(circuit, param.tolist(), f.tolist())
    z1d = np.zeros(Z.size*2, dtype=np.float64)
    z1d[0:z1d.size:2] = err.real
    z1d[1:z1d.size:2] = err.imag
    if valid(circuit, param):
        return z1d
    else:
        return 1e6*np.ones(Z.size*2, dtype=np.float64)


def valid(circuit, param):
    """ checks validity of parameters

    Parameters
    ----------
    circuit : string
        string defining the circuit

    param : list
        list of parameter values

    Returns
    -------
    valid : boolean

    Notes
    -----
    All parameters are considered valid if they are greater than zero --
    except for E2 (the exponent of CPE) which also must be less than one.

    """

    p_string = [x for x in circuit if x not in 'ps(),-/']

    for i, (a, b) in enumerate(zip(p_string[::2], p_string[1::2])):
        if str(a + b) == "E2":
            if param[i] <= 0 or param[i] >= 1:
                return False
        else:
            if param[i] <= 0:
                return False

    return True


def computeCircuit(circuit, parameters, frequencies):
    """ evaluates a circuit string for a given set of parameters and frequencies

    Parameters
    ----------
    circuit : string
    parameters : list of floats
    frequencies : list of floats

    Returns
    -------
    array of floats
    """

    return eval(buildCircuit(circuit, parameters, frequencies))


def buildCircuit(circuit, parameters, frequencies):
    """ transforms a circuit, parameters, and frequencies into a string
    that can be evaluated

    Parameters
    ----------
    circuit : str
    parameters : list of floats
    frequencies : list of floats

    Returns
    -------
    eval_string : str
        Python expression for calculating the resulting fit
    """

    series_string = "s(["
    for elem in circuit.split("-"):
        element_string = ""
        if "p" in elem:
            parallel_string = "p(("
            for par in elem.strip("p()").split(","):
                param_string = ""
                elem_type = par[0]
                elem_number = len(par.split("/"))

                param_string += str(parameters[0:elem_number])
                parameters = parameters[elem_number:]

                new_elem = (elem_type + "(" + param_string + "," +
                                        str(frequencies) + "),")
                parallel_string += new_elem

            element_string = parallel_string.strip(",") + "))"
        else:
            param_string = ""
            elem_type = elem[0]
            elem_number = len(elem.split("/"))

            param_string += str(parameters[0:elem_number])
            parameters = parameters[elem_number:]

            element_string = (elem_type + "(" + param_string + "," +
                                          str(frequencies) + ")")

        series_string += element_string + ","

    eval_string = series_string.strip(",") + "])"

    return eval_string


def calculateCircuitLength(circuit):
    l1 = ['R', 'E', 'W', 'C', 'A', 'G']
    length = 0
    for char in l1:
        length += circuit.count(char)
    return length



from .fitting import circuit_fit, computeCircuit, calculateCircuitLength
from .plotting import plot_nyquist
import matplotlib.pyplot as plt
import numpy as np


class BaseCircuit:
    """ Base class for equivalent circuit models """
    def __init__(self, initial_guess=None, name=None,
                 algorithm='leastsq', bounds=None):
        """ Base constructor for any equivalent circuit model """

        # if supplied, check that initial_guess is valid and store
        if initial_guess is not None:
            for i in initial_guess:
                assert isinstance(i, (float, int, np.int32, np.float64)),\
                    'value {} in initial_guess is not a number'.format(i)

        # initalize class attributes
        self.initial_guess = initial_guess
        self.name = name
        self.algorithm = algorithm
        self.bounds = bounds

        # initialize fit parameters and confidence intervals
        self.parameters_ = None
        self.conf_ = None

    def fit(self, frequencies, impedance):
        """ Fit the circuit model

        Parameters
        ----------
        frequencies: numpy array
            Frequencies

        impedance: numpy array of dtype 'complex128'
            Impedance values to fit

        Returns
        -------
        self: returns an instance of self

        """

        # check that inputs are valid:
        #    frequencies: array of numbers
        #    impedance: array of complex numbers
        #    impedance and frequency match in length

        assert isinstance(frequencies, np.ndarray),\
            'frequencies is not of type np.ndarray'
        assert isinstance(frequencies[0], (float, int, np.int32, np.float64)),\
            'frequencies does not contain a number'
        assert isinstance(impedance, np.ndarray),\
            'impedance is not of type np.ndarray'
        assert isinstance(impedance[0], (complex, np.complex128)),\
            'impedance does not contain complex numbers'
        assert len(frequencies) == len(impedance),\
            'mismatch in length of input frequencies and impedances'

        if self.initial_guess is not None:
            parameters, conf = circuit_fit(frequencies, impedance,
                                           self.circuit, self.initial_guess,
                                           self.algorithm,
                                           bounds=self.bounds)
            self.parameters_ = parameters
            if conf is not None:
                self.conf_ = conf
        else:
            # TODO auto calc guess
            raise ValueError('no initial guess supplied')

        return self

    def _is_fit(self):
        """ check if model has been fit (parameters_ is not None) """
        if self.parameters_ is not None:
            return True
        else:
            return False

    def predict(self, frequencies):
        """ Predict impedance using a fit equivalent circuit model

        Parameters
        ----------
        frequencies: numpy array
            Frequencies

        Returns
        -------
        impedance: numpy array of dtype 'complex128'
            Predicted impedance

        """

        # check that inputs are valid:
        #    frequencies: array of numbers

        assert isinstance(frequencies, np.ndarray),\
            'frequencies is not of type np.ndarray'
        assert isinstance(frequencies[0], (float, int, np.int32, np.float64)),\
            'frequencies does not contain a number'

        if self._is_fit():
            return computeCircuit(self.circuit,
                                  self.parameters_.tolist(),
                                  frequencies.tolist())

        else:
            raise ValueError("The model hasn't been fit yet. " +
                             "Please call the `.fit` method before trying to" +
                             " predict model output")

    def __str__(self):
        """ Defines the pretty printing of the circuit """

        # parse the element names from the circuit string
        names = self.circuit.replace('p', '').replace('(', '').replace(')', '')
        names = names.replace(',', '-').replace('/', '-').split('-')

        to_print  = '\n-------------------------------\n'  # noqa E222
        to_print += 'Circuit: {}\n'.format(self.name)
        to_print += 'Circuit string: {}\n'.format(self.circuit)
        to_print += 'Algorithm: {}\n'.format(self.algorithm)

        if self._is_fit():
            to_print += 'Fit: True\n'
            to_print += 'Fit parameters:\n'
            for name, param in zip(names, self.parameters_):
                to_print += '\t{} = {:.2e}\n'.format(name, param)
        else:
            to_print += 'Fit: False\n'
            to_print += 'Initial guesses:\n'
            for name, param in zip(names, self.initial_guess):
                to_print += '\t{} = {:.2e}\n'.format(name, param)

        to_print += '\n-------------------------------\n'
        return to_print

    def plot(self, f_data=None, Z_data=None, CI=True):
        """ a convenience method for plotting Nyquist plots


        Parameters
        ----------
        f_data: np.array of type float
            Frequencies of input data (for Bode plots)
        Z_data: np.array of type complex
            Impedance data to plot
        CI: boolean
            Include bootstrapped confidence intervals in plot

        Returns
        -------
        ax: matplotlib.axes
            axes of the created nyquist plot
        """

        fig, ax = plt.subplots(figsize=(5, 5))

        if Z_data is not None:
            ax = plot_nyquist(ax, f_data, Z_data)

        if self._is_fit():

            if f_data is not None:
                f_pred = f_data
            else:
                f_pred = np.logspace(5, -3)

            Z_fit = self.predict(f_pred)
            ax = plot_nyquist(ax, f_data, Z_fit, fit=True)

            if CI:
                N = 1000
                n = len(self.parameters_)
                f_pred = np.logspace(np.log10(min(f_data)),
                                     np.log10(max(f_data)),
                                     num=100)

                params = self.parameters_
                confs = self.conf_

                full_range = np.ndarray(shape=(N, len(f_pred)), dtype=complex)
                for i in range(N):
                    self.parameters_ = params + \
                                        confs*np.random.uniform(-2, 2, size=n)

                    full_range[i, :] = self.predict(f_pred)

                self.parameters_ = params

                min_Z = []
                max_Z = []
                for x in np.real(Z_fit):
                    ys = []
                    for run in full_range:
                        ind = np.argmin(np.abs(run.real - x))
                        ys.append(run[ind].imag)

                    min_Z.append(x + 1j*min(ys))
                    max_Z.append(x + 1j*max(ys))

                ax.fill_between(np.real(min_Z), -np.imag(min_Z),
                                -np.imag(max_Z), alpha='.2')

        plt.show()


class Randles(BaseCircuit):
    """ A Randles circuit model class """
    def __init__(self, CPE=False, **kwargs):
        """ Constructor for the Randles' circuit class

        Parameters
        ----------
        CPE: boolean
            Use a constant phase element instead of a capacitor
        """
        super().__init__(**kwargs)

        if CPE:
            self.name = 'Randles w/ CPE'
            self.circuit = 'R_0-p(R_1,E_1/E_2)-W_1/W_2'
            circuit_length = calculateCircuitLength(self.circuit)
            assert len(self.initial_guess) == circuit_length, \
                'Initial guess length needs to be equal to parameter length'
        else:
            self.name = 'Randles'
            self.circuit = 'R_0-p(R_1,C_1)-W_1/W_2'
            circuit_length = calculateCircuitLength(self.circuit)
            assert len(self.initial_guess) == circuit_length, \
                'Initial guess length needs to be equal to parameter length'


class CustomCircuit(BaseCircuit):
    def __init__(self, circuit=None, **kwargs):
        """ Constructor for a customizable equivalent circuit model

        Parameters
        ----------
        circuit: string
            A string that should be interpreted as an equivalent circuit

        """

        super().__init__(**kwargs)
        self.circuit = circuit

        circuit_length = calculateCircuitLength(self.circuit)
        assert len(self.initial_guess) == circuit_length, \
            'Initial guess length needs to be equal to {circuit_length}'


class FlexiCircuit(BaseCircuit):
    def __init__(self, max_elements=None, generations=2,
                 popsize=30, initial_guess=None):
        """ Constructor for the Flexible Circuit class

        Parameters
        ----------
        max_elements: integer
            The maximum number of elements available to the algorithm
        solve_time: integer
            The maximum allowed solve time, in seconds.

        """

        self.name = 'Flexible Circuit'
        self.initial_guess = initial_guess
        self.generations = generations
        self.popsize = popsize
        self.max_elements = max_elements
#        self.bounds = bounds

    def fit(self, frequencies, impedances):
        from scipy.optimize import leastsq
        from .genetic import make_population
        from .fitting import residuals
        n = 5
        f = frequencies
        Z = impedances
        for i in range(self.generations):
            scores = []
            self.population = make_population(self.popsize, n)
            for pop in self.population:
                self.circuit = pop
                print(self.circuit)
                circuit_length = calculateCircuitLength(self.circuit)
#                for j in range(n-4,n+4):
#                    print(residuals(self.initial_guess,Z,f,self.circuit))
#                    try:
#                        print(self.circuit)
                self.initial_guess = list(np.ones(circuit_length)/10)
#                        print(residuals(self.initial_guess,Z,F,))
#                        print(self.initial_guess)
                p_values, covar, ff, _, _ = leastsq(residuals,
                                                    self.initial_guess,
                                                    args=(Z, f, self.circuit),
                                                    maxfev=100000, ftol=1E-13,
                                                    full_output=True)

                print(p_values)
                scores.append([np.square(ff['fvec']).mean(), pop])

        print(scores)
        return scores
